#include "seahaven.h"

struct HashBucket {
  struct {
    Stack s;
    int y;		// hack, really want position w/in stack as an index
  } cardpos[52];
  HashBucket *next;
};

static const int NUMHASHSLOTS = 337;

HashBucket *hashtable[NUMHASHSLOTS] = {NULL};

void
InitHashTable()
{
    for (int i=0; i< NUMHASHSLOTS; i++) {
	if (hashtable[i]) {
	    HashBucket *t = hashtable[i], *n;
	    do {
		n = t->next;
		delete t;
	    } while (t = n);
	    hashtable[i] = NULL;
	}
    }
}

#ifdef notdef
// the single stacks are symmetric, so store them as singlestack[0] and y=0
inline void FoldStack(Stack reals, int realy, Stack &s, int &y)
{
	extern Stack singlestack[];
	if (reals == singlestack[0] || reals == singlestack[1]
		|| reals == singlestack[2] || reals == singlestack[3]) {
	  s = singlestack[0];
	  y = 0;
	} else {
	  s = reals;
	  y = realy;
	}
}
#endif

// retruns true if found it
bool
hashlookup(int key)
{
    if (!hashtable[key]) return false;

    HashBucket *t = hashtable[key];
    do {
	int i;
	Card *c;
	bool match;

	for (i=0, c = &cards[0][0]; i < 52; i++, c++) {
	    match = true;
	    if ( (*c)->getFoldedY() != t->cardpos[i].y ||
		(*c)->getFoldedStack() != t->cardpos[i].s ) {
		match = false;
		break;
	    }
	}
	if (match)
	    return true;
    } while (t = t->next);

    return false;

}

// returns true if new state
bool
CheckNewState()
{
    unsigned int hashkey = 0;

    // compute key
    int i;
    Card *c;
    for (i=0, c = &cards[0][0]; i < 52; i++, c++) {
	hashkey += ((int) (*c)->getFoldedStack() << (*c)->getValue());
    }
    
    if (hashlookup(hashkey % NUMHASHSLOTS))
	return false;
    else
	return true;
    
}

void
hashinsert(int key, HashBucket *pail)
{
    pail->next = hashtable[key];
    hashtable[key] = pail;
}

void
SaveState()
{
    
    HashBucket *pail = new HashBucket;
    unsigned int hashkey = 0;
    
    int i;
    Card *c;
    for (i=0, c = &cards[0][0]; i < 52; i++, c++) {
	pail->cardpos[i].s = (*c)->getFoldedStack();
	pail->cardpos[i].y = (*c)->getFoldedY();
	hashkey += ((int) pail->cardpos[i].s << (*c)->getValue());
    }
    
    hashinsert(hashkey % NUMHASHSLOTS, pail);
    
}
